/*
 * freetrans - main.c
 * 	version 0.0.3
 *
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <freetrans.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <comm.h>
#include <time.h>
#include <signal.h>
#include <inttypes.h>
#include <tread.h>
#include <display.h>
#include <pthread.h>
#include <sys/wait.h>
#include <ar2c.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <ar2c_parser.h>
#include <sys/time.h>
#include <crc16.h>

#define TT_FILE		1
#define TT_SOCKET	2
#define TT_FIFO		3
#define TT_SERIAL	4

#define ST_NO_TIME	0
#define ST_ABS_TIME	1
#define ST_REL_TIME	2
#define ST_SEND_TIME	3

#define CRC16_LE	1
#define CRC16_BE	2

static int targ_type = TT_FILE;
static char *targ_opt = NULL;
static char *targ_opt2 = NULL;
static int port = 2000;		/* additional parameter for TCP/UDP */
static int ifd, ofd;
static int silent_level = 1;
static speed_t bspeed = B9600;
static int ser_par = COMM_PAR_EVEN;
static int ser_stop = COMM_PAR_CSTOPB;
static int speed = 0, rspeed = 0;
static int send_wait = 1;
static unsigned int recv_cols = 0;
static unsigned long long show_time = ST_NO_TIME;
static unsigned long long send_time = 0;
static unsigned long repeat_count = 1;
static unsigned long repeat_delay = 0;
static int add_crc = 0;
static int ascii = 0;
static int use_udp;

static int want_quit;

#define PROMPT	":"

static char *rl_gets(void)
{
	char *tmp;

	tmp = readline(PROMPT);
	if (tmp && *tmp)
		add_history(tmp);
	return tmp;
}

static unsigned long long time_us(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return (tv.tv_sec * 1000000ULL + tv.tv_usec);
}

static ssize_t my_write(int fd, const void *buf, size_t len)
{
	static void *my_buffer;
	static size_t my_buffer_size;
	unsigned long i;
	uint16_t crc;
	
	switch (add_crc) {
	case CRC16_LE:
	case CRC16_BE:
		if (my_buffer_size < len + 2) {
			free(my_buffer);
			my_buffer_size = 2 * (len + 2);
			my_buffer = malloc(my_buffer_size);
			if (my_buffer == NULL)
				return -ENOMEM;
		}

		crc = crc16(0xffff, buf, len);
		if (add_crc == CRC16_BE) {
			crc = cpu_to_be16(crc);
		} else {
			crc = cpu_to_le16(crc);
		}
		memcpy(my_buffer, buf, len);
		memcpy((char *) my_buffer + len, &crc, 2);

		buf = my_buffer;
		len += 2;
		break;
	}

	if (show_time == ST_SEND_TIME)
		send_time = time_us();
	write(fd, buf, len);

	for (i = 1; i < repeat_count; i++) {
		if (repeat_delay)
			usleep(repeat_delay);
		write(fd, buf, len);
	}
	return len;
}

static void show_usage(char *name)
{
	fprintf(stderr,	"usage: %s [options]\n"
			"Options:\n"
			"  -A, --ascii		ASCII mode\n"
			"  -B, --bbitrate	base bitrate(default 9600bps)\n"
			"  -b, --bitrate\t	set bitrate\n"
			"  -d, --device		use selected device\n"
			"  -f, --fifo		use dual fifo(for wr)\n"
			"  -h, --help		show this help\n"
			"      --parity		set parity (none, odd, even)\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"      --stopbits	stop bits (1/2)\n"
			"  -u, --udp		use UDP\n"
			"  -v, --verbose		be verbose\n"
			"  -V, --version		show version\n",
			(name == NULL) ? "freetrans" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"ascii", 0, 0, 'A'},
		{"bbitrate", 1, 0, 'B'},
		{"bitrate", 1, 0, 'b'},
		{"datazone", 0, 0, 'z'},
		{"device", 1, 0, 'd'},
		{"fifo", 1, 0, 'f'},
		{"help", 0, 0, 'h'},
		{"host", 1, 0, 's'},
		{"parity", 1, 0, 3},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"rbitrate", 1, 0, 2},
		{"stopbits", 1, 0, 4},
		{"udp", 0, 0, 'u'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "AB:b:d:f:p:hs:vquVOz",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
			case 2:
				rspeed = itoar2cspeed(atoi(optarg));
				if (rspeed < 0) {
					fprintf(stderr, "Invalid return speed "
								"setting.");
					exit(EXIT_FAILURE);
				}
				targ_type = TT_SERIAL;
				break;
			case 3:
				if (!strcmp(optarg, "none"))
					ser_par = COMM_PAR_NONE;
				else if (!strcmp(optarg, "odd"))
					ser_par = COMM_PAR_ODD;
				else if (!strcmp(optarg, "even"))
					ser_par = COMM_PAR_EVEN;
				else {
					fprintf(stderr, "Invalid parity: %s.\n",
							optarg);
					exit(EXIT_FAILURE);
				}
				break;
			case 4:
				if (!strcmp(optarg, "1"))
					ser_stop = 0;
				else if (!strcmp(optarg, "2"))
					ser_stop = COMM_PAR_CSTOPB;
				else {
					fprintf(stderr, "Invalid stop bits: %s."
							"\n", optarg);
					exit(EXIT_FAILURE);
				}
				break;
			case 'A':
				ascii = 1;
				break;
			case 'B':
				bspeed = itospeed(atoi(optarg));
				if (bspeed < 0) {
					fprintf(stderr, "Unsupported initial "
							"speed setting.\n");
					exit(EXIT_FAILURE);
				}
				if (rspeed == 0) {
					rspeed = speedtoar2cspeed(bspeed);
					if (rspeed < 0)
						rspeed = 0;
				}
				targ_type = TT_SERIAL;
				break;
			case 'b':
				speed = itoar2cspeed(atoi(optarg));
				if (speed < 0) {
					fprintf(stderr, "Invalid speed "
							"setting.");
					exit(EXIT_FAILURE);
				}
				targ_type = TT_SERIAL;
				break;
			case 'd':
				targ_opt = optarg;
				if (targ_type == TT_SOCKET)
					targ_type = TT_FILE;
				break;
			case 'f':
				targ_opt2 = optarg;
				targ_type = TT_FIFO;
				break;
			case 'h':
				show_usage(args[0]);
				break;
			case 'p':
				port = atoi(optarg);
				targ_type = TT_SOCKET;
				break;
			case 'v':
				silent_level++;
				break;
			case 'q':
				silent_level--;
				break;
			case 's':
				targ_opt = optarg;
				targ_type = TT_SOCKET;
				break;
			case 'u':
				use_udp = 1;
				break;
			case 'V':
				show_version();
				exit(EXIT_SUCCESS);
				break;
			default:
				fprintf(stderr, "Unknown option.\n");
				exit(EXIT_FAILURE);
		}
	}
	if (optind < argc) {
		fprintf(stderr, "warning: unknown parameters detected.\n");
	}
}

static pthread_mutex_t display_lock = PTHREAD_MUTEX_INITIALIZER;

static inline void disp_lock(void)
{
	pthread_mutex_lock(&display_lock);
}

static inline void disp_unlock(void)
{
	pthread_mutex_unlock(&display_lock);
}

#define RBUFF_SIZE	256

/*
 * FIXME: this is ugly
 */
volatile int rl_stop = 0;

static void *read_loop(void *args)
{
	ssize_t ret, i;
	uint8_t buff[RBUFF_SIZE];
	int f = 0;
	int col = 0;
	int cols;

	while (!rl_stop) {
		ret = t_read_nb(ifd, buff, RBUFF_SIZE, 200);
		if (ret < 0) {
			rl_stop = 2;
			return NULL;
		}
		if (ascii) {
			for (i = 0; i < ret; i++)
				if (buff[i] != '\r')
					wprintw(rw, "%c", (int) buff[i]);
		} else if (show_time) {
			if (ret == 0)
				continue;
			wprintw(rw, "%f: ", (float) (time_us() - send_time)
					/ 1000000.0);
			for (i = 0; i < ret; i++)
				wprintw(rw, "%02x ", buff[i]);
			wprintw(rw, "\n");
		} else {
			if (ret == 0) {
				if (f) {
					wprintw(rw, "\n");
					col = 0;
					f = 0;
				}
				continue;
			}
			f = 1;
       			cols = recv_cols ? recv_cols : COLS / 3;
			for (i = 0; i < ret; i++) {
				if (++col > cols) {
					wprintw(rw, "\n");
					col = 1;
				}
				wprintw(rw, "%02x ", buff[i]);
			}
		}
		disp_lock();
		wrefresh(rw);
		disp_unlock();
	}
	rl_stop = 0;
	return NULL;
}

pthread_t read_thread;

static int read_thread_init(void)
{
	if (pthread_create(&read_thread, NULL, read_loop, NULL))
		fatal("Can't create thread.\n");
	pthread_detach(read_thread);
	return 0;
}

static void show_id(void)
{
	int id;

	id = ar2c_get_id(ifd, ofd);
	if (id < 0)
		fprintf(stderr, "get_id error!\n");
	else
		printf("remote id = %d\n", id);
}

#define ADD_FLAG(name, bit)						\
	do {								\
		if ((flags & (1 << (bit)))) {				\
			if (!first)					\
				strncat(buf, ",", len);			\
			else						\
				first = 0;				\
			strncat(buf, (name), len);			\
		}							\
	} while (0)

static void flags2str(char *buf, size_t len, uint32_t flags)
{
	int first = 1;
	buf[0] = 0;
	ADD_FLAG("wysrej", 0);
	ADD_FLAG("setspd", 1);
	ADD_FLAG("modesw", 2);
	ADD_FLAG("prnctl", 3);
	ADD_FLAG("rejinfo", 4);
	ADD_FLAG("cfset", 5);
	ADD_FLAG("ngm", 6);
	ADD_FLAG("wydal", 7);
	ADD_FLAG("crypto", 8);
}

static void ar2c_lp_info_show(struct ar2c_lp_info *info)
{
	char sfs[128], cfs[128];
	int i;

	flags2str(sfs, 128, info->supported_flags);
	flags2str(cfs, 128, info->current_flags);
	printf("AR-2c LP info extenstions supported by remote side:\n"
			"Protocol version:\t%hhi\n"
			"Model:\t\t\t%hhi\n"
			"Software version:\t0x%08x\n"
			"Supported flags:\t0x%08x (%.60s)\n"
			"Current flags:\t\t0x%08x (%.60s)\n"
			"Description:\t\t%.48s\n",
			info->protocol_version, info->model_id,
			info->software_version, info->supported_flags, sfs,
			info->current_flags, cfs, info->description);

	if ((info->current_flags & AR2C_LPIF_SPDSET) &&
			(targ_type == TT_SERIAL)) {
		printf("Supported serial speed: ");
		for (i = 0; i < ar2c_spd_count(ifd, ofd); i++)
			printf("%s ", ar2c_spd_name(i));
		printf("\n");
	}
}

static int ar2c_lp_info(void)
{
	uint8_t cmd[] = { 0xff, 0x0f, 0x15, 0x00 };
	struct ar2c_lp_info resp;
	int retry = 2, ret;

	while (retry-- > 0) {
		write(ofd, cmd, sizeof(cmd));
		ret = t_read(ifd, (char *) &resp, sizeof(struct ar2c_lp_info),
				250);
		if (ret < sizeof(struct ar2c_lp_info)) {
			do_resync(ifd, ofd);
			continue;
		}
		ar2c_lp_info_show(&resp);
		return 0;
	}
	return 1;
}

static void cmd_setspd(const char *args)
{
	unsigned long spd;
	speed_t ser_spd;
	char cmd[64];

	if (targ_type != TT_SERIAL) {
		fprintf(stderr, "Setting speed is supported only on serial "
				"links.\n");
		return;
	}
	sscanf(args, "%63s %li", cmd, &spd);
	ser_spd = itospeed(spd);
	if (ser_spd == -EINVAL) {
		fprintf(stderr, "Invalid speed setting.\n");
		return;
	}
	serial_set_speed(ifd, ser_spd, ser_par | ser_stop);
}

static void cmd_ar2csetspd(const char *args)
{
	unsigned long spd;
	int ser_spd;
	char cmd[64];

	if (targ_type != TT_SERIAL) {
		fprintf(stderr, "Setting speed is supported only on serial "
				"links.\n");
		return;
	}
	sscanf(args, "%63s %li", cmd, &spd);
	ser_spd = itoar2cspeed(spd);
	if (ser_spd < 0) {
		fprintf(stderr, "Invalid speed setting.\n");
		return;
	}
	ar2c_set_speed(ifd, ser_spd, ser_par | ser_stop);
}

static void set_recv_cols(const char *args)
{
	char cmd[64];

	sscanf(args, "%63s %i", cmd, &recv_cols);
}

static void set_repeat_count(const char *args)
{
	char cmd[64];

	sscanf(args, "%63s %li", cmd, &repeat_count);
}

#define SEND_BUFFLEN	1024

struct send_buffer {
	unsigned int obs;
	uint8_t buff[SEND_BUFFLEN];
};

static void sb_flush(struct send_buffer *sb)
{
	my_write(ofd, sb->buff, sb->obs);
	sb->obs = 0;
}

/*
 * positive len are big-endian, negative are little-endian
 */
static int send_word(struct send_buffer *sb, unsigned int w, int len)
{
	if (sb->obs + abs(len) >= SEND_BUFFLEN)
		sb_flush(sb);

	switch (len) {
		case 1:
		case -1:
			sb->buff[sb->obs++] = w;
			break;
		case 2:
			*((uint16_t *) &sb->buff[sb->obs]) = cpu_to_be16(w);
			sb->obs += 2;
			break;
		case -2:
			*((uint16_t *) &sb->buff[sb->obs]) = cpu_to_le16(w);
			sb->obs += 2;
			break;
		case 4:
			*((uint32_t *) &sb->buff[sb->obs]) = cpu_to_be32(w);
			sb->obs += 4;
			break;
		case -4:
			*((uint32_t *) &sb->buff[sb->obs]) = cpu_to_le32(w);
			sb->obs += 4;
			break;
		default:
			return -EINVAL;
	}
	return 0;
}

static void send_message(const char *args)
{
	struct send_buffer sb;
	unsigned int state = 0, tmp = 0, len = 1;
	char c;

	while ((c = *args++)) {
		if (c == ' ')
			break;
	}
	if (c == '\0')
		return;

	sb.obs = 0;
	for (;;) {
		c = *args++;

		if (c == '\0') {
			if (state)
				send_word(&sb, tmp, len);
			break;
		}

		switch (state) {
			case 0: /* none */
				if (c == '0')
					state = 1;
				else if (isdigit(c)) {
					state = 3;
					tmp = c - '0';
				} else switch (c) {
					case ' ':
					case '\t':
						break;
					case 'W':
						len = 2;
						break;
					case 'w':
						len = -2;
						break;
					case 'L':
						len = 4;
						break;
					case 'l':
						len = -4;
						break;
					case '-':
						state = 5;
						break;
					default:
						fprintf(stderr, "parse error."
								"\n");
				}
				break;
			case 1: /* 0 */
				if (c == 'x')
					state = 4;
				else if (c >= '0' && c <= '7') {
					tmp = c - '0';
					state = 2;
				} else {
					send_word(&sb, tmp, len);
					state = 0;
					len = 1;
					tmp = 0;
				}
				break;
			case 2: /* 0[0-9]+ */
				if (c >= '0' && c <= '7')
					tmp = (tmp << 3) | (c - '0');
				else {
					send_word(&sb, tmp, len);
					state = 0;
					len = 1;
					tmp = 0;
				}
				break;
			case 3: /* [1-9] */
				if (isdigit(c))
					tmp = 10 * tmp + c - '0';
				else {
					send_word(&sb, tmp, len);
					state = 0;
					len = 1;
					tmp = 0;
				}
				break;
			case 4: /* 0x */
				if (isdigit(c))
					tmp = (tmp << 4) + c - '0';
				else if (c >= 'a' && c <= 'f')
					tmp = (tmp << 4) + c - 'a' + 10;
				else if (c >= 'A' && c <= 'F')
					tmp = (tmp << 4) + c - 'A' + 10;
				else {
					send_word(&sb, tmp, len);
					state = 0;
					len = 1;
					tmp = 0;
				}
				break;
			case 5: /* -[1-9] */
				if (isdigit(c))
					tmp = 10 * tmp - c + '0';
				else {
					send_word(&sb, tmp, len);
					state = 0;
					len = 1;
					tmp = 0;
				}
				break;
			default:
				BUG();
		}
	}
	sb_flush(&sb);
}

static void set_repeat_delay(const char *args)
{
	char cmd[64];

	sscanf(args, "%63s %li", cmd, &repeat_delay);
}

static void set_crc(const char *args)
{
	char cmd[64], type[64];

	if (sscanf(args, "%63s %63s", cmd, type) < 2)
		fprintf(stderr, "set crc needs an argument.\n");
	else if (!strcmp(type, "none"))
		add_crc = 0;
	else if (!strcmp(type, "crc16") || !strcmp(type, "crc16le"))
		add_crc = CRC16_LE;
	else if (!strcmp(type, "crc16be"))
		add_crc = CRC16_BE;
	else
		fprintf(stderr, "unknown crc type\n");
}

static void cmd_open(const char *args)
{
	char cmd[64], file[256];
	int fd;
	int ret;
	
	ret = sscanf(args, "%63s %255s", cmd, file);
	if (ret < 2) {
		fprintf(stderr, "usage: open FILE\n");
		return;
	}
	file[255] = '\0';

	fd = open(file, O_RDWR);
	if (fd == -1) {
		fprintf(stderr, "open %s: %s\n", args, strerror(errno));
		return;
	}

	if (ifd != ofd)
		close(ifd);
	close(ofd);
	ifd = ofd = fd;
}

static void cmd_connect(const char *args)
{
	char cmd[64], host[64], proto[64];
	int fd;
	int ret;

	ret = sscanf(args, "%63s %63s %d %63s", cmd, host, &port, proto);
	if (ret < 3) {
		fprintf(stderr, "usage: connect HOST PORT [PROTOCOL]\n");
		return;
	}
	host[63] = '\0';
	proto[63] = '\0';

	use_udp = 0;
	if (ret == 4) {
		if (!strcmp(proto, "tcp")) {
			use_udp = 0;
		} else if (!strcmp(proto, "udp")) {
			use_udp = 1;
		} else {
			fprintf(stderr, "connect: unknown protocol %s\n",
					proto);
			return;
		}
	}
	fd = pcomm_open(host, port, use_udp);
	if (fd == -1) {
		fprintf(stderr, "error: Can't connect to %s port %d\n", host,
				port);
		return;
	}
	if (ifd != ofd)
		close(ifd);
	close(ofd);

	fprintf(stderr, "connect: connected fd=%d\n", fd);

	ifd = ofd = fd;
}

static void cmd_version(void)
{
	fprintf(stderr,
"freetrans " PACKAGE_VERSION "\n"
"Copyright (C) 2005..2009  Krzysztof Mazur <krzysiek@podlesie.net>\n\n");
	show_version();
}

static void cmd_help(void)
{
	fprintf(stderr,
"freetrans commands:\n"
"	!			execute shell command with remote connection\n"
"	abstime			show timestamps (0 = 01.01.1970)\n"
"	ar2csetspd		AR-2c LP set serial speed (ext.)\n"
"	ascii			ASCII mode\n"
"	connect			connect to HOST PORT [PROTOCOL]\n"
"	crc MODE		set checksuming mode (none, crc16le, crc16be)\n"
"	exit, quit, q		terminate freetrans\n"
"	?, help			show this help\n"
"	hex			HEX mode\n"
"	id			get remote AR-2c LP id\n"
"	info			get remote AR-2c LP info (ext.)\n"
"	notime			normal mode (no timestamps)\n"
"	open			connect to specified FILE\n"
"	rcols, recv_cols	recive window columns\n"
"	return, ret, r		return to send mode\n"
"	repeat_count, rcount	set number of repetitions of command\n"
"	repeat_delay, rdelay	set delay between repeated commands in us\n"
"	send, s			send message\n"
"	setspd, setspeed	set serial speed\n"
"	stime			show timestamps (0 = send time)\n"
"	time			show timestamps (0 = now)\n"
"	ver, version		show version\n"
"	wait/nowait		wait for <ENTER> ?\n"
		);
}

static int subcommand_pid = -1;

static void shell_subcommand(const char *tmp)
{
	const char *t;
	int ret;
	pid_t pid;
	int status;

	if (*tmp != '!')
		return;

	for (t = tmp + 1; whitespace(*t); t++)
		;
	if (*t == '\0')
		return;

	pid = fork();
	if (pid == -1) {
		fprintf(stderr, "file() failed: %s\n", strerror(errno));
		return;
	}

	if (!pid) {
		close(STDIN_FILENO);
		close(STDOUT_FILENO);

		ret = dup2(ifd, STDIN_FILENO);
		if (ret == -1) {
			fprintf(stderr, "dup failed: %s\n", strerror(errno));
			_exit(1);
		}

		ret = dup2(ofd, STDOUT_FILENO);
		if (ret == -1) {
			fprintf(stderr, "dup out failed: %s\n",
					strerror(errno));
			_exit(1);
		}

		if (ifd != ofd)
			close(ofd);
		close(ifd);

		execl("/bin/sh", "sh", "-c", t, NULL);
		/* child */
		_exit(0);
	}
	subcommand_pid = pid;
	waitpid(pid, &status, 0);
	subcommand_pid = -1;
	return;
}

static void parse_command(void)
{
	char cmd[64];
	char *tmp = NULL;

	/* FIXME */
	if (rl_stop != 2) {
		rl_stop = 1;
		do {
			sched_yield();
		} while (rl_stop);
	} else {
		rl_stop = 0;
	}

	endwin();
	while (1) {
		if (tmp)
			free(tmp);
		tmp = rl_gets();
		if (tmp == NULL) {
			fprintf(stderr, "\n");
			break;
		}
		if (*tmp == '\0')
			break;
		if (sscanf(tmp, "%63s", cmd) < 1)
			continue;
		if (!strcmp(cmd, "exit") || !strcmp(cmd, "quit") ||
				!strcmp(cmd, "q"))
			exit(EXIT_SUCCESS);
		if (!strcmp(cmd, "return") || !strcmp(cmd, "ret") ||
				!strcmp(cmd, "r"))
			break;
		if (cmd[0] == '!') {
			shell_subcommand(tmp);
			continue;
		}
		if (!strcmp(cmd, "id")) {
			show_id();
			continue;
		}
		if (!strcmp(cmd, "info")) {
			ar2c_lp_info();
			continue;
		}
		if (!strcmp(cmd, "wait")) {
			send_wait = 1;
			continue;
		}
		if (!strcmp(cmd, "nowait")) {
			send_wait = 0;
			continue;
		}
		if (!strcmp(cmd, "setspd") || !strcmp(cmd, "setspeed")) {
			cmd_setspd(tmp);
			continue;
		}
		if (!strcmp(cmd, "ar2csetspd") ||
				!strcmp(cmd, "ar2csetspeed")) {
			cmd_ar2csetspd(tmp);
			continue;
		}
		if (!strcmp(cmd, "rcols") || !strcmp(cmd, "recv_cols")) {
			set_recv_cols(tmp);;
			continue;
		}
		if (!strcmp(cmd, "?") || !strcmp(cmd, "help")) {
			cmd_help();
			continue;
		}
		if (!strcmp(cmd, "time")) {
			show_time = ST_REL_TIME;
			send_time = time_us();
			continue;
		}
		if (!strcmp(cmd, "stime")) {
			show_time = ST_SEND_TIME;
			send_time = time_us();
			continue;
		}
		if (!strcmp(cmd, "abstime")) {
			show_time = ST_ABS_TIME;
			continue;
		}
		if (!strcmp(cmd, "notime")) {
			show_time = ST_NO_TIME;
			continue;
		}
		if (!strcmp(cmd, "version") || !strcmp(cmd, "ver")) {
			cmd_version();
			continue;
		}
		if (!strcmp(cmd, "repeat_count") || !strcmp(cmd, "rcount")) {
			set_repeat_count(tmp);
			continue;
		}
		if (!strcmp(cmd, "repeat_delay") || !strcmp(cmd, "rdelay")) {
			set_repeat_delay(tmp);
			continue;
		}
		if (!strcmp(cmd, "crc")) {
			set_crc(tmp);
			continue;
		}
		if (!strcmp(cmd, "send") || !strcmp(cmd, "s")) {
			send_message(tmp);
			continue;
		}
		if (!strcmp(cmd, "open")) {
			cmd_open(tmp);
			continue;
		}
		if (!strcmp(cmd, "connect")) {
			cmd_connect(tmp);
			continue;
		}
		if (!strcmp(cmd, "x")) {
			refresh();
			read_thread_init();
			send_message(tmp);
			if (tmp != NULL)
				free(tmp);
			return;
		}
		if (!strcmp(cmd, "ascii")) {
			ascii = 1;
			continue;
		}
		if (!strcmp(cmd, "hex")) {
			ascii = 0;
			continue;
		}
		fprintf(stderr, "Unknown command: %s\n", cmd);
	}
	if (tmp != NULL)
		free(tmp);
	refresh();
	read_thread_init();
}

#define BUFF_SIZE	4096
static int main_loop(void)
{
	uint8_t buff[BUFF_SIZE];
	int off, f, c;

	off = 0;
	f = 0;
	while (1) {
		c = wgetch(sw);
		if (want_quit) {
			if (subcommand_pid != -1) {
				kill(subcommand_pid, SIGINT);
				subcommand_pid = -1;
				want_quit = 0;
			} else {
				break;
			}
		}
		if (c == ERR)
			continue;
		if (ascii) {
			char t = c;

			if (c == ':') {
				parse_command();
				continue;
			}

			my_write(ofd, &t, 1);
			wprintw(sw, "%c", c);
			disp_lock();
			wrefresh(sw);
			disp_unlock();
		} else if (c >= '0' && c <= '9') {
			if (f) {
				if (off < BUFF_SIZE) {
					buff[off] |= c - '0';
					if (!send_wait)
						my_write(ofd, &buff[off], 1);
					wprintw(sw, "%c ", c);
					disp_lock();
					wrefresh(sw);
					disp_unlock();
					f = 0;
					off++;
				}
			} else {
				buff[off] = (c - '0') << 4;
				wprintw(sw, "%c", c);
				disp_lock();
				wrefresh(sw);
				disp_unlock();
				f = 1;
			}
		} else if (c >= 'a' && c <= 'f') {
			if (f) {
				if (off < BUFF_SIZE) {
					buff[off] |= c + 10 - 'a';
					if (!send_wait)
						my_write(ofd, &buff[off], 1);
					wprintw(sw, "%c ", c);
					disp_lock();
					wrefresh(sw);
					disp_unlock();
					f = 0;
					off++;
				}
			} else {
				wprintw(sw, "%c", c);
				disp_lock();
				wrefresh(sw);
				disp_unlock();
				buff[off] = (c + 10 - 'a') << 4;
				f = 1;
			}
		} else switch (c) {
			case 0xd:
				wprintw(sw, "\n");
				if (send_wait)
					my_write(ofd, buff, off);
				disp_lock();
				wrefresh(sw);
				disp_unlock();
				off = 0;
				f = 0;
				break;
			case ':':
				parse_command();
				break;
			case 'Q':
				rl_stop = 1;
				do {
					sched_yield();
				} while (rl_stop);
				endwin();
				exit(EXIT_SUCCESS);
				break;
			case KEY_BACKSPACE:
				if (f) {
					f = 0;
					wprintw(sw, "\b \b");
				} else if (off > 0) {
					f = 1;
					off--;
					buff[off] &= 0xf0;
					wprintw(sw, "\b\b  \b\b");
				}
				disp_lock();
				wrefresh(sw);
				disp_unlock();
				break;
			default:
				break;
		}
	}
}

/**
 * restore_serial - restore serial settings
 *
 * FIXME: this version restores only remote speed settings, I cound
 * be a good idea to restore also local serial settings.
 */
static void restore_serial(void)
{
	if (speed) {
		if (rspeed == 0) {
			rspeed = speedtoar2cspeed(bspeed);
			if (rspeed < 0) {
				fprintf(stderr, "Unable to return to initial "
						"serial speed.\n");
				exit(EXIT_FAILURE);
			}
		}
		if ((speed != rspeed) && ar2c_set_speed(ifd, rspeed,
					ser_par | ser_stop)) {
			fprintf(stderr, "Setting serial speed error\n");
			exit(EXIT_FAILURE);
		}
	}
}

static char *history_file = NULL;

void init_history(void)
{
	char *home = getenv("HOME");

	using_history();
	stifle_history(1000);
	history_file = malloc(strlen(home) + 100);
	if (history_file != NULL) {
		sprintf(history_file, "%s/.freetrans_history", home);
		read_history(history_file);
	}
}

static void prepare_exit(void)
{
	if (targ_type == TT_FILE)
		restore_serial();

	if (history_file != NULL)
		write_history(history_file);	
}

static const char *commands[] = {
	"!",
	"abstime",
	"ar2csetspeed",
	"ascii",
	"connect",
	"crc",
	"exit",
	"help",
	"hex",
	"id",
	"info",
	"notime",
	"nowait",
	"open",
	"quit",
	"rcols",
	"rcount",
	"rdelay",
	"recv_cols",
	"repeat_count",
	"repeat_delay",
	"return",
	"setspeed",
	"stime",
	"time",
	"version",
	"wait",
	NULL
};

static char *command_generator(const char *text, int state)
{
	static int list_index;
	static int len;
	const char *name;

	if (!state) {
		list_index = 0;
		len = strlen(text);
	}

	while ((name = commands[list_index]) != NULL) {
		list_index++;

		if (!strncmp(name, text, len))
			return strdup(name);
	}

	return NULL;
}

static char **freetrans_completion(const char *text, int start, int end)
{
	char **matches;

	matches = NULL;

	if (start == 0)
		matches = rl_completion_matches(text, command_generator);
	return matches;
}

static void initialize_readline(void)
{
	rl_readline_name = "freetrans";

	rl_attempted_completion_function = freetrans_completion;
}

static void sigint_handler(int sig)
{
	want_quit = 1;
}

static struct sigaction sigint_action = {
	.sa_flags		= SA_RESTART,
};

int main(int argc, char **args)
{
	fprintf(stderr, "freetrans " PACKAGE_VERSION ".\n"
			"Copyright (C) 2005..2010  Krzysztof Mazur.\n");

	sigint_action.sa_handler = sigint_handler;

	signal(SIGALRM, SIG_IGN);
	sigaction(SIGINT, &sigint_action, NULL);
	
	parse_command_line(argc, args);
	switch (targ_type) {
		case TT_FILE:
			if (targ_opt == NULL)
				targ_opt = strdup("/dev/null");

			if ((ifd = open(targ_opt, O_RDWR)) < 0) {
				fprintf(stderr, "error: Can't open %s:%s\n",
						targ_opt, strerror(errno));
				exit(EXIT_FAILURE);
			}
			ofd = ifd;
			break;
		case TT_SOCKET:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(EXIT_FAILURE);
			}
			if ((ifd = pcomm_open(targ_opt, port, use_udp)) < 0) {
				fprintf(stderr, "error: Can't connect to %s\n",
						targ_opt);
				exit(EXIT_FAILURE);
			}
			ofd = ifd;
			break;
		case TT_FIFO:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(EXIT_FAILURE);
			}
			if (targ_opt2 == NULL) {
				fprintf(stderr, "fatal: null targ_opt2.\n");
				exit(EXIT_FAILURE);
			}
			/* order is important */
			if ((ofd = open(targ_opt2, O_WRONLY)) == -1) {
				fprintf(stderr, "fatal: Can't open wr_fifo.\n");
				exit(EXIT_FAILURE);
			}
			if ((ifd = open(targ_opt, O_RDONLY)) == -1) {
				fprintf(stderr, "fatal: Can't open rd_fifo.\n");
				exit(EXIT_FAILURE);
			}
			break;
		case TT_SERIAL:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(EXIT_FAILURE);
			}
			if ((ifd = open(targ_opt, O_RDWR)) < 0) {
				fprintf(stderr, "error: Can't open %s: %s\n",
						targ_opt, strerror(errno));
				exit(EXIT_FAILURE);
			}
			ofd = ifd;
			serial_set_speed(ifd, bspeed, ser_par | ser_stop);
			if (ar2cspeedtospeed(speed) != bspeed)
				if (ar2c_set_speed(ifd, speed, ser_par
							| ser_stop)) {
					fprintf(stderr, "Setting serial speed "
							"error\n");
					exit(EXIT_FAILURE);
				}
			break;
		default:
			fprintf(stderr, "fatal: Internal error.\n");
			exit(EXIT_FAILURE);
	}

	initialize_readline();
	init_history();
	atexit(prepare_exit);

	ftd_init();
	read_thread_init();
	main_loop();
	endwin();
	return 0;
}
